require "src/vector"
require "src/triangle"
require "src/mesh"
require "src/matrix"

function love.load()
    ---------------------------setup screen size-----------------------------------------
    screen = {}
    screen.dpiScale = love.graphics.getDPIScale()
    screen.width, screen.height = 800, 600
    screen.width, screen.height = screen.width / screen.dpiScale, screen.height / screen.dpiScale
    --love.graphics.setWireframe(true)
    --love.window.setMode(screen.width, screen.height)
    --love.window.setTitle("3D Renderer")
    

    cube = {}
    cube.triangles =  LoadMesh("meshes/1701.obj")
    
    Camera = Vector.new(0, 0, 0)
    Light_dir = Vector.new(0, 0, 0.5)

    ----------------------------Projection Matrix---------------------------------------------
    near = 0.1
    far = 1000
    fov = 90 
    aspectRatio = screen.height / screen.width
    fovRad = 1 / math.tan(fov*0.5/180* math.pi)
    Theta = 0
   
    
end

function LoadMesh(filename)
    local triangles = {}
    local lines = {}
    local start = love.timer.getTime()
    for line in love.filesystem.lines(filename) do 
        lines[#lines + 1] = line
    end
    for i=1, 3 do --get rid of unnessecary header
         table.remove(lines, 1)
    end
    --load vector
    local j = 1
    local vector = {}
    local c = 0
    local stop = false
    while stop == false do 
        vector[j] = {}
        for num in string.gmatch(lines[j],"%S+") do
            if num ~= "v"  then
                if num == "s" then 
                    stop = true
                end
                if stop == false then 
                    if c + 1 < 4 then 
                        c = c + 1
                    else c = 1 end
                    if c == 1 then 
                        vector[j].x = tonumber(num)
                    elseif c == 2 then 
                        vector[j].y = tonumber(num)
                    elseif c == 3 then 
                        vector[j].z = tonumber(num)
                    end
                            
                end
            end
        end
        if stop == false then 
            j = j+1
        end
    end

    --put together vector into traingles / faces
    local m = {}
    c=1
    for i=#vector+1, #lines do 
        for num in string.gmatch(lines[i],"%S+") do
            if num ~= "f"  then 
                table.insert( m,tonumber( num))
            end
        end
    end
   
    for i=1, #lines-#vector do 

        triangles[i] = Triangle.new({vector[m[c]], vector[m[c+1]], vector[m[c+2]]})
        c = c +3 
    end



    local stop =   love.timer.getTime()
    -- test 
    for i=1, 26 do 
        --print(i, vector[i].x, vector[i].y, vector[i].z )
    end
    print("Total loading Time:" .. stop - start)
    print("Total triangles: " .. #triangles)
    return triangles
end

function love.update(dt)
    love.window.setTitle(love.timer.getFPS())
    matProj = Matrix.new(4, 4)
    matRotX = Matrix.new(4,4)
    matRotZ = Matrix.new(4,4)

    Theta = Theta + 1 *dt

    matRotZ.m[1][1] = math.cos(Theta)
    matRotZ.m[1][2] = math.sin(Theta)
    matRotZ.m[2][1] = -math.sin(Theta)
    matRotZ.m[2][2] = math.cos(Theta)
    matRotZ.m[3][3] = 1
    matRotZ.m[4][4] = 1

    matRotX.m[1][1] = 1 
    matRotX.m[2][2] = math.cos(Theta* 0.5)
    matRotX.m[2][3] = math.sin(Theta* 0.5)
    matRotX.m[3][2] = -math.sin(Theta* 0.5)
    matRotX.m[3][3] = math.cos(Theta * 0.5)
    matRotX.m[4][4] = 1

    matProj.m[1][1] = aspectRatio * fovRad
    matProj.m[2][2] = fovRad
    matProj.m[3][3] = far / (far - near)
    matProj.m[4][3] = (-far * near) / (far - near)
    matProj.m[3][4] = 1
    matProj.m[4][4] = 0
end

function love.draw()
    local start = love.timer.getTime()
    for i, tris in ipairs(cube.triangles) do
        local triProjected = Triangle.new({Vector.new(0,0,0), Vector.new(0,0,0), Vector.new(0,0,0)} )
        local triTranslated = Triangle.new({Vector.new(0,0,0), Vector.new(0,0,0), Vector.new(0,0,0)} )
        local triRotatedZ = Triangle.new({Vector.new(0,0,0), Vector.new(0,0,0), Vector.new(0,0,0)} )
        local triRotatedX = Triangle.new({Vector.new(0,0,0), Vector.new(0,0,0), Vector.new(0,0,0)} )

        --rotate
        triRotatedZ.vectors[1] = MultiplyMatrixVector(tris.vectors[1], matRotZ) 
        triRotatedZ.vectors[2] = MultiplyMatrixVector(tris.vectors[2], matRotZ) 
        triRotatedZ.vectors[3] = MultiplyMatrixVector(tris.vectors[3], matRotZ)

        triRotatedX.vectors[1] = MultiplyMatrixVector(triRotatedZ.vectors[1], matRotX) 
        triRotatedX.vectors[2] = MultiplyMatrixVector(triRotatedZ.vectors[2], matRotX) 
        triRotatedX.vectors[3] = MultiplyMatrixVector(triRotatedZ.vectors[3], matRotX)

        --push into view
        triTranslated.vectors[1] = Vector.new(triRotatedX.vectors[1].x, triRotatedX.vectors[1].y, triRotatedX.vectors[1].z+1)
        triTranslated.vectors[2] = Vector.new(triRotatedX.vectors[2].x, triRotatedX.vectors[2].y, triRotatedX.vectors[2].z+1)
        triTranslated.vectors[3] = Vector.new(triRotatedX.vectors[3].x, triRotatedX.vectors[3].y, triRotatedX.vectors[3].z+1)

        --culling
        local normal = Vector.new(0, 0, 0); local line1 = Vector.new(0, 0, 0); local line2 = Vector.new(0, 0, 0);

        line1.x = triTranslated.vectors[2].x - triTranslated.vectors[1].x;
        line1.y = triTranslated.vectors[2].y - triTranslated.vectors[1].y;
        line1.z = triTranslated.vectors[2].z - triTranslated.vectors[1].z;

        line2.x = triTranslated.vectors[3].x - triTranslated.vectors[1].x;
        line2.y = triTranslated.vectors[3].y - triTranslated.vectors[1].y;
        line2.z = triTranslated.vectors[3].z - triTranslated.vectors[1].z;

        normal.x = line1.y * line2.z - line1.z * line2.y;
        normal.y = line1.z * line2.x - line1.x * line2.z;
        normal.z = line1.x * line2.y - line1.y * line2.x;
        
        normal = NormalizeVec3d(normal)
        
        if  DotProduct(normal, triTranslated.vectors[1], Camera) < 0 then
            --lighting
            --Light_dir = NormalizeVec3d(Light_dir)
            if DotProduct(normal, Light_dir) < 0 then 
                d = DotProduct(normal, Light_dir)
                d = d*2
                love.graphics.setColor(d*-1, d*-1, d*-1)
            end

            --Projection from 3D to 2D
            triProjected.vectors[1] = MultiplyMatrixVector(triTranslated.vectors[1], matProj);
            triProjected.vectors[2] = MultiplyMatrixVector(triTranslated.vectors[2], matProj);
            triProjected.vectors[3] = MultiplyMatrixVector(triTranslated.vectors[3], matProj);


            --scale into view
            triProjected.vectors[1].x = triProjected.vectors[1].x+1;  triProjected.vectors[1].y = triProjected.vectors[1].y+1;
            triProjected.vectors[2].x = triProjected.vectors[2].x+1;  triProjected.vectors[2].y = triProjected.vectors[2].y+1;
            triProjected.vectors[3].x = triProjected.vectors[3].x+1;  triProjected.vectors[3].y = triProjected.vectors[3].y+1;

            triProjected.vectors[1].x = triProjected.vectors[1].x * 0.5 * screen.width;
            triProjected.vectors[1].y = triProjected.vectors[1].y * 0.5 * screen.height;
            triProjected.vectors[2].x = triProjected.vectors[2].x * 0.5 * screen.width;
            triProjected.vectors[2].y = triProjected.vectors[2].y * 0.5 * screen.height;
            triProjected.vectors[3].x = triProjected.vectors[3].x * 0.5 * screen.width;
            triProjected.vectors[3].y = triProjected.vectors[3].y * 0.5 * screen.height;

            --solid shaded scolor 
            --love.graphics.polygon("fill", 
            --lriProjected.vectors[1].x, triProjected.vectors[1].y,
            --lriProjected.vectors[2].x, triProjected.vectors[2].y, 
            --lriProjected.vectors[3].x, triProjected.vectors[3].y)


            --this is the problem, if i comment this out there are no visual glitches, but i need it for debugging
            love.graphics.setColor(1, 1, 1)
            love.graphics.polygon("line", 
            triProjected.vectors[1].x, triProjected.vectors[1].y,
            triProjected.vectors[2].x, triProjected.vectors[2].y, 
            triProjected.vectors[3].x, triProjected.vectors[3].y)

        end
    end
    local stop = love.timer.getTime()
    love.graphics.setColor(1, 1, 0)
    love.graphics.print("Frametime:" .. stop- start)


end

function MultiplyMatrixVector(i,matrix)
    local o = {x=0, y=0, z=0}
    o.x = i.x * matrix.m[1][1] + i.y * matrix.m[2][1] + i.z * matrix.m[3][1] + matrix.m[4][1]
    o.y = i.x * matrix.m[1][2] + i.y * matrix.m[2][2] + i.z * matrix.m[3][2] + matrix.m[4][2]
    o.z = i.x * matrix.m[1][3] + i.y * matrix.m[2][3] + i.z * matrix.m[3][3] + matrix.m[4][3]
    local w = i.x * matrix.m[1][4] + i.y * matrix.m[2][4] + i.z * matrix.m[3][4] + matrix.m[4][4]
    if  w ~= 0 then
        o.x = o.x / w 
        o.y = o.y / w 
        o.z = o.z / w
    end
    return o
end

function NormalizeVec3d(vector)
    local lenght = math.sqrt(vector.x*vector.x + vector.y*vector.y + vector.z*vector.z);
    return{ x = vector.x / lenght, y = vector.y / lenght, z = vector.z / lenght}
end

function DotProduct(vec, vec2, vec3)
    if vec3 == nil then 
        vec3 = Vector.new(0, 0, 0)
    end
    return vec.x * (vec2.x - vec3.x)+ vec.y * (vec2.y - vec3.y)+ vec.z * (vec2.z - vec3.z)
end